﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.IO;

namespace NBtest
{
    public class Row
    {
        public int nClusterIndex = -1;
        public List<int> m_files = new List<int>(0);
        public List<int> m_data = new List<int>(0);
        public List<int> m_pos = new List<int>(0);
        public double norm = 0.0;
    }

    class NB
    {
        private int[] m_fileCat = null;
        private int[] m_files = null;
        private int[] m_tabooList = null;
        private string[] m_fileName = null;
        public List<Row> m_rows = new List<Row>();

        private int m_nWords = -1;
        private float[][] logP = null;
        private double[] m_sizeCoeff = null;

        public void readOnlyTrainingSet(string datafile, int nLow, int nHigh)
        {
            m_rows.Clear();
            FileInfo fi = new FileInfo(datafile);
            long nBytes = fi.Length;
            int nNonZeros = (int)(nBytes / 10);
            int nRold = -1;
            int current_row = -1;
            bool isStarted = false;
            using (FileStream stream = new FileStream(datafile, FileMode.Open))
            {
                using (BinaryReader reader = new BinaryReader(stream))
                {
                    for (long i = 0; i < nNonZeros; ++i)
                    {
                        int nR = reader.ReadInt32();
                        int nC = reader.ReadInt32();
                        short sum = reader.ReadInt16();
                        if (nC > m_nWords) m_nWords = nC;
                        if (nR != nRold)
                        {
                            bool isFound = false;
                            int index = 0;
                            foreach (int k in m_files)
                            {
                                if (k == nR)
                                {
                                    isFound = true;
                                    break;
                                }
                                ++index;
                            }
                            if (isFound)
                            {
                                if (m_tabooList[index] >= nLow && m_tabooList[index] <= nHigh)
                                {
                                    m_rows.Add(new Row());
                                    ++current_row;
                                    m_rows[current_row].m_files.Add(nR);
                                    if (m_rows[current_row].nClusterIndex < 0)
                                    {
                                        m_rows[current_row].nClusterIndex = m_tabooList[index];
                                    }
                                    isStarted = true;
                                }
                            }
                            else
                            {
                                isStarted = false;
                            }
                            nRold = nR;
                        }
                        else
                        {
                            if (isStarted)
                            {
                                m_rows[current_row].m_pos.Add(nC);
                                m_rows[current_row].m_data.Add(sum);
                            }
                        }
                    }
                    reader.Close();
                }
                stream.Close();
            }
            ++m_nWords;
        }

        public void readAndProcessAllData(string datafile, int nLow, int nHigh)
        {
            FileInfo fi = new FileInfo(datafile);
            long nBytes = fi.Length;
            int nNonZeros = (int)(nBytes / 10);
            int nRold = -1;
            int nR = 0;
            bool isStarted = false;
            Row row = new Row();
            int cnt = 0;
            using (FileStream stream = new FileStream(datafile, FileMode.Open))
            {
                using (BinaryReader reader = new BinaryReader(stream))
                {
                    for (long i = 0; i < nNonZeros; ++i)
                    {
                        nR = reader.ReadInt32();
                        int nC = reader.ReadInt32();
                        short sum = reader.ReadInt16();
                        if (nR != nRold)
                        {
                            bool isFound = false;
                            int index = 0;
                            foreach (int k in m_files)
                            {
                                if (k == nR)
                                {
                                    isFound = true;
                                    break;
                                }
                                ++index;
                            }
                            if (isFound)
                            {
                                if (m_tabooList[index] >= nLow && m_tabooList[index] <= nHigh)
                                {
                                    if (row.m_data.Count > 0)
                                    {
                                        m_fileCat[nR] = getCategory(row);
                                        ++cnt;
                                    }

                                    row.m_data.Clear();
                                    row.m_files.Clear();
                                    row.m_pos.Clear();

                                    row.m_files.Add(nR);
                                    if (row.nClusterIndex < 0)
                                    {
                                        row.nClusterIndex = m_tabooList[index];
                                    }
                                    isStarted = true;
                                }
                            }
                            else
                            {
                                isStarted = false;
                            }
                            nRold = nR;
                        }
                        else
                        {
                            if (isStarted)
                            {
                                row.m_pos.Add(nC);
                                row.m_data.Add(sum);
                            }
                        }
                    }
                    reader.Close();
                }
                stream.Close();
            }

            if (row.m_data.Count > 0)
            {
                m_fileCat[nR] = getCategory(row);
                ++cnt;
            }
        }

        public int getNumberOfTrainedClusters()
        {
            return m_rows.Count;
        }

        private bool parseFileListString(string s, ref int index, ref int taboo, ref string name)
        {
            if (s == null) return false;
            string[] local = s.Split(' ');
            if (local == null) return false;
            if (local.Length < 3) return false;

            try
            {
                index = Convert.ToInt32(local[0].Trim());
                taboo = Convert.ToInt32(local[1].Trim());
                name = local[2].Trim();
            }
            catch (Exception e)
            {
                Console.WriteLine(e.Message);
                Environment.Exit(1);
            }

            return true;
        }

        public bool populateFileAndTabooLists(string listFiles)
        {
            if (!File.Exists(listFiles)) return false;

            int index = 0;
            int taboo = -1;
            string name = "";
            List<int> listIndex = new List<int>();
            List<int> listTaboo = new List<int>();
            List<string> listName = new List<string>();
            StreamReader fileListStream = new StreamReader(listFiles);
            string s = string.Empty;
            while (s != null)
            {
                s = fileListStream.ReadLine();
                if (!parseFileListString(s, ref index, ref taboo, ref name))
                {
                    break;
                }
                else
                {
                    listIndex.Add(index);
                    listTaboo.Add(taboo);
                    listName.Add(name);
                }
            }
            fileListStream.Close();

            int size = listIndex.Count;
            m_files = new int[size];
            m_tabooList = new int[size];
            m_fileName = new string[size];
            m_fileCat = new int[size];

            for (int i = 0; i < size; ++i)
            {
                m_files[i] = listIndex[i];
                m_tabooList[i] = listTaboo[i];
                m_fileName[i] = listName[i];
            }
            listIndex.Clear();
            listTaboo.Clear();
            listName.Clear();

            return true;
        }

        private Row mergeTwoRows(Row r1, Row r2)
        {
            Row merged = new Row();

            for (int i = 0; i < r1.m_files.Count; ++i)
            {
                merged.m_files.Add(r1.m_files[i]);
            }
            for (int i = 0; i < r2.m_files.Count; ++i)
            {
                merged.m_files.Add(r2.m_files[i]);
            }

            int index1 = 0;
            int index2 = 0;
            while (true)
            {
                if (r1.m_pos[index1] == r2.m_pos[index2])
                {
                    merged.m_pos.Add(r1.m_pos[index1]);
                    int sum = r1.m_data[index1];
                    if (r2.m_data[index2] > r1.m_data[index1])
                    {
                        sum = r2.m_data[index2];
                    }
                    merged.m_data.Add(sum);
                    ++index1;
                    ++index2;
                }
                else if (r1.m_pos[index1] > r2.m_pos[index2])
                {
                    merged.m_pos.Add(r2.m_pos[index2]);
                    merged.m_data.Add(r2.m_data[index2]);
                    ++index2;
                }
                else if (r1.m_pos[index1] < r2.m_pos[index2])
                {
                    merged.m_pos.Add(r1.m_pos[index1]);
                    merged.m_data.Add(r1.m_data[index1]);
                    ++index1;
                }
                if (index1 >= r1.m_pos.Count) break;
                if (index2 >= r2.m_pos.Count) break;
            }

            for (int i = index1; i < r1.m_pos.Count; ++i)
            {
                merged.m_pos.Add(r1.m_pos[i]);
                merged.m_data.Add(r1.m_data[i]);
            }

            for (int i = index2; i < r2.m_pos.Count; ++i)
            {
                merged.m_pos.Add(r2.m_pos[i]);
                merged.m_data.Add(r2.m_data[i]);
            }

            if (r1.nClusterIndex >= 0 && r2.nClusterIndex >= 0)
            {
                if (r1.nClusterIndex != r2.nClusterIndex)
                {
                    Console.WriteLine("Error: merging two knowingly different clusters {0} {1}", r1.nClusterIndex, r2.nClusterIndex);
                    Environment.Exit(0);
                }
            }

            if (r1.nClusterIndex >= 0) merged.nClusterIndex = r1.nClusterIndex;
            if (r2.nClusterIndex >= 0) merged.nClusterIndex = r2.nClusterIndex;

            return merged;
        }

        public void showClusters()
        {
            int N = m_rows.Count;
            for (int i = 0; i < N; ++i)
            {
                Console.WriteLine(" {0} {1}", m_rows[i].nClusterIndex, m_rows[i].m_files[0]);
            }
            Console.WriteLine();
        }

        public int getNumberOfClusters()
        {
            return m_rows.Count;
        }

        public bool mergeKnownClusters()
        {
            int N = m_rows.Count;
            bool wasMerged = false;
            for (int i = 0; i < N; ++i)
            {
                if (m_rows[i].nClusterIndex >= 0)
                {
                    Row r1 = m_rows[i];
                    for (int j = i + 1; j < N; ++j)
                    {
                        if (m_rows[i].nClusterIndex == m_rows[j].nClusterIndex)
                        {
                            Row r2 = m_rows[j];
                            Row row = mergeTwoRows(r1, r2);
                            m_rows.Remove(r1);
                            m_rows.Remove(r2);
                            m_rows.Add(row);
                            wasMerged = true;
                            break;
                        }
                    }
                    if (wasMerged) break;
                }
            }
            if (wasMerged)
            {
                mergeKnownClusters();
            }
            return true;
        }

        public void showFilesInClusters()
        {
            Console.WriteLine("Training sample\n");
            int N = m_rows.Count;
            for (int i = 0; i < N; ++i)
            {
                for (int j = 0; j < m_rows[i].m_files.Count; ++j)
                {
                    Console.Write(" {0} ", m_rows[i].m_files[j]);
                }
                Console.WriteLine();
            }
            Console.WriteLine();
        }

        public void computeCoeff()
        {
            int N = m_rows.Count;
            m_sizeCoeff = new double[N];
            for (int i = 0; i < N; ++i)
            {
                long size = 0;
                for (int j = 0; j < m_rows[i].m_files.Count; ++j)
                {
                    
                    int index = m_rows[i].m_files[j];
                    //Console.WriteLine("*{0}*", m_fileName[index]);
                    FileInfo fi = new FileInfo(m_fileName[index]);
                    size += fi.Length;
                }
                m_sizeCoeff[i] = (double)(size);
                m_sizeCoeff[i] = Math.Log(m_sizeCoeff[i]);
            }
        }

        private int getMaxElement(double[] array, int size)
        {
            int pos = 0;
            double max = array[pos];
            for (int i = 1; i < size; ++i)
            {
                if (max < array[i])
                {
                    max = array[i];
                    pos = i;
                }
            }
            return pos;
        }

        private int getCategory(Row row)
        {
            int nCat = m_rows.Count;
            double[] categories = new double[nCat];
            for (int i = 0; i < nCat; ++i)
            {
                categories[i] = 0.0;
                for (int j = 0; j < row.m_data.Count; ++j)
                {
                    categories[i] += logP[i][row.m_pos[j]] * row.m_data[j];
                }
            }

            for (int i = 0; i < nCat; ++i)
            {
                //this is computational trick a coefficient that takes into consideration
                //the size of the training file set, it can improve accuracy on about 20%
                categories[i] *= m_sizeCoeff[i];
            }
            return getMaxElement(categories, nCat);
        }

        public void buildMatrix()
        {
            int nDocs = m_rows.Count;
            logP = new float[nDocs][];
            for (int i = 0; i < nDocs; ++i)
            {
                logP[i] = new float[m_nWords];
            }
            for (int i = 0; i < nDocs; ++i)
            {
                for (int j = 0; j < m_nWords; ++j)
                {
                    logP[i][j] = 1.0f;
                }
            }
            int cnt = 0;
            foreach (Row row in m_rows)
            {
                for (int i = 0; i < row.m_data.Count; ++i)
                {
                    logP[cnt][row.m_pos[i]] += (float)(row.m_data[i]);
                }
                ++cnt;
            }

            for (int i = 0; i < m_nWords; ++i)
            {
                float total = 0.0f;
                for (int j = 0; j < nDocs; ++j)
                {
                    total += logP[j][i];
                }
                for (int j = 0; j < nDocs; ++j)
                {
                    logP[j][i] /= total;
                }
            }

            for (int i = 0; i < m_nWords; ++i)
            {
                for (int j = 0; j < nDocs; ++j)
                {
                    logP[j][i] = (float)(Math.Log((logP[j][i])));
                }
            }
        }

        public void showActualCategories()
        {
            foreach (int x in m_fileCat)
            {
                Console.WriteLine(x);
            }
        }

        public double computeF_measure(int[] nExpected, int nCategories)
        {
            int[,] nAccuracy = new int[nCategories, nCategories];
            for (int i = 0; i < nCategories; ++i)
            {
                for (int j = 0; j < nCategories; ++j)
                {
                    nAccuracy[i, j] = 0;
                }
            }


            for (int i = 0; i < nCategories; ++i)
            {
                for (int j = 0; j < nExpected.Length; ++j)
                {
                    if (nExpected[j] == i)
                    {
                        ++nAccuracy[i, m_fileCat[j]];
                    }
                }
            }

            Console.WriteLine("Table of recognized categories\n");
            for (int i = 0; i < nCategories; ++i)
            {
                for (int j = 0; j < nCategories; ++j)
                {
                    Console.Write(" {0,4} ", nAccuracy[i, j]);
                }
                Console.WriteLine();
            }
            Console.WriteLine();

            int[] sumInRows = new int[nCategories];
            int[] sumInCols = new int[nCategories];
            for (int i = 0; i < nCategories; ++i)
            {
                sumInRows[i] = 0;
                sumInCols[i] = 0;
                for (int j = 0; j < nCategories; ++j)
                {
                    sumInRows[i] += nAccuracy[i, j];
                    sumInCols[i] += nAccuracy[j, i];
                }
            }

            double[,] f_measure = new double[nCategories, nCategories];
            for (int i = 0; i < nCategories; ++i)
            {
                for (int j = 0; j < nCategories; ++j)
                {
                    if (nAccuracy[i, j] > 0)
                    {
                        double f = (double)nAccuracy[i, j];
                        double prec = f / (double)(sumInRows[i]);
                        double recall = f / (double)(sumInCols[j]);

                        f_measure[i, j] = 2.0 * prec * recall / (prec + recall);
                    }
                    else
                    {
                        f_measure[i, j] = 0.0;
                    }
                }
            }

            double total = 0.0;
            for (int loop = 0; loop < nCategories; ++loop)
            {
                int RR = 0;
                int CC = 0;
                double max = f_measure[RR, CC];
                for (int i = 0; i < nCategories; ++i)
                {
                    for (int j = 0; j < nCategories; ++j)
                    {
                        if (f_measure[i, j] > max)
                        {
                            max = f_measure[i, j];
                            RR = i;
                            CC = j;
                        }
                    }
                }
                total += max;
                for (int i = 0; i < nCategories; ++i)
                {
                    f_measure[RR, i] = 0.0;
                }
                for (int j = 0; j < nCategories; ++j)
                {
                    f_measure[j, CC] = 0.0;
                }
            }
            return total / (double)(nCategories);
        }
    }
}
