﻿using System;
using System.Collections.Generic;
using System.Collections;
using System.Linq;
using System.Text;
using System.IO;
using System.Reflection;
using wordprocessing;

namespace NBtest
{
    class Program
    {
        static void Main(string[] args)
        {
            string rootFolder = @"..//..//..//..//PATENTCORPUS128";
            string extension = "*.txt";
            string dataFolder = "data";
            string filesList = "ProcessedFilesList.txt";
            string dataFile = "DocWordMatrix.dat";
            string path = Path.GetDirectoryName(Assembly.GetExecutingAssembly().Location);
            dataFolder = Path.Combine(path, dataFolder);
            filesList = Path.Combine(dataFolder, filesList);
            dataFile = Path.Combine(dataFolder, dataFile);

            DirectoryInfo di = new DirectoryInfo(rootFolder);
            if (!di.Exists)
            {
                Console.WriteLine("The data folder not found, please correct the path");
                return;
            }

            DateTime start = DateTime.Now;
            //
            //This is training sample, it is good only PATENTCORPUS128.
            //For other data training sample can be set in file ProcessedFilesList.txt
            //Example of two lines from ProcessedFilesList.txt:
            //3 0 C:\APOLAR\UTILITIES\PATENTCORPUS128\C399P7058343.txt
            //4 -1 C:\APOLAR\UTILITIES\PATENTCORPUS128\C399P7164866.txt
            //first number is index of processed file, second number is known category if >= 0 or
            //indicator that file to be categorized if < 0;
            //Make sure the full file name has no spaces or change parsing function.
            //
            int N = 128;
            int[] nChart = new int[N];
            for (int i = 0; i < N; ++i)
            {
                nChart[i] = -1;
            }
            nChart[0] = 0; nChart[1] = 0;  
            nChart[16] = 1; nChart[17] = 1; nChart[18] = 1; 
            nChart[32] = 2;  
            nChart[48] = 3;   
            nChart[64] = 4;  
            nChart[80] = 5; nChart[81] = 5;
            nChart[96] = 6; 
            nChart[112] = 7; 
            //
            WordProcessing wp = new WordProcessing(dataFolder, filesList, dataFile);
            wp.processFiles(rootFolder, extension, nChart);

            //Following is Naive Bayes processing
            int nCategories = 8;  //that is good for PATENTCORPUS128.

            NB nb = new NB();
            nb.populateFileAndTabooLists(filesList);
            nb.readOnlyTrainingSet(dataFile, 0, nCategories);
            int nClusters = nb.getNumberOfClusters();
            if (nClusters <= 0)
            {
                Console.WriteLine("Training sample is not specified");
                return;
            }

            //nb.showClusters();
            if (!nb.mergeKnownClusters())
            {
                Console.WriteLine("Error merging known categories");
                return;
            }
            nb.showFilesInClusters();
            nb.computeCoeff();
            nb.buildMatrix();
            nb.readAndProcessAllData(dataFile, -1, nCategories);
            //nb.showActualCategories();

            //This is expected result. It is necessary only for estimation of accuracy
            //This data is good only for PATENTCORPUS128.
            int[] nExpected = new int[128];
            int nCatSize = 16;
            int nWhich = 0;
            for (int i = 0; i < 128; ++i)
            {
                nExpected[i] = nWhich;
                if ((i + 1) % nCatSize == 0)
                {
                    ++nWhich;
                }
            }
            Console.WriteLine("Accuracy {0}", nb.computeF_measure(nExpected, nCategories));
 
            DateTime end = DateTime.Now;
            TimeSpan duration = end - start;
            double time = duration.Minutes * 60.0 + duration.Seconds + duration.Milliseconds / 1000.0;
            Console.WriteLine("Total processing time {0:########.00} seconds", time);
        }
    }
}
